/*
** -----------------------------------------------------------------------------**
** pixproc.v
**
** Processes pixel data - buffers and combines outputs from two halves of the CCD
** (the CCD readout on both outputs starts from the outer and progresses to the middle of each line),
** adds pixel level shifts and scales (for each color)
**
** Copyright (C) 2003-2007 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x347
**  x347 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: pixproc.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.5  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/
// color/gain correction, windowing
// will remove second output later - for now it will be just optimized out as it is not used in the top module

module pixproc (wclk, // interface clock
                weg,   // write enable (and cs)
                wew,   // write enable (and cs)
                we_ll,// write half-line-length (LSB - 1 pixel shift). Adjust to merge 2 halves 
                a,    // [3:0] register address 0..3 - gains channel 1, 4..7 - channel 2, 8..b/c..f - left, width,top, height
                d,    // [11:0] gains - 8 bits, sizes - 12 bits
                aclk, // ADC/pixel clock
                clk2x, // output clock, twice ADC, 
                frame, // high during frame 
                line, // high during pixel shift 
                da,   // [13:0] data in, channel A
                db,   // [13:0] data in, channel B
                qa,   // [13:0] data out, channel A
//                qb,   // [13:0] data out, channel B
//                hact_a, hact_b, // valid pixel output
//                vact_a, vact_b);// active frame output
                hact_a, // valid pixel output
                vact_a);// active frame output
//  input         wclk, we, we_ll, aclk, clk2x, frame,line;
  input         wclk, weg, wew, we_ll, aclk, clk2x, frame,line;
  input  [ 3:0] a;
  input  [15:0] d;
  input  [13:0] da;
  input  [13:0] db;
  output [13:0] qa;
//  output [13:0] qb;
//  output        hact_a, hact_b;
//  output        vact_a, vact_b;
  output        hact_a;
  output        vact_a;


  reg  line_d;
  reg oddpix, oddline;
  wire   [15:0] gain_a;
  wire   [15:0] gain_b;
  wire   [13:0] sub_a;
  wire   [13:0] sub_b;
  reg    [13:0] dca;
  reg    [13:0] dcb;
  reg    [13:0] dsa;
  reg    [13:0] dsb;
  wire   [29:0] dm_a=dsa[13:0]*gain_a[15:0];
  wire   [29:0] dm_b=dsb[13:0]*gain_b[15:0];


  wire   [13:0] do;

  wire   double_line; // merged data active
  reg    [10:0] ll;

  wire   [14:0] dsub_a = {1'b0,da[13:0]} - sub_a[13:0];
  wire   [14:0] dsub_b = {1'b0,db[13:0]} - sub_b[13:0];

// add registers if needed
  wire   [17:0] dm_round_a=dm_a[29:12]+dm_a[11];
  wire   [17:0] dm_round_b=dm_b[29:12]+dm_b[11];

  always @ (posedge aclk) begin
    dsa[13:0] <=dsub_a[14]?14'b0:dsub_a[13:0];
    dsb[13:0] <=dsub_b[14]?14'b0:dsub_b[13:0];
    dca[13:0] <=(dm_round_a[17:14]==4'b0)?(dm_round_a[13:0]):14'h3fff;
    dcb[13:0] <=(dm_round_b[17:14]==4'b0)?(dm_round_b[13:0]):14'h3fff;
  end
  assign        qa[13:0]= hact_a? do[13:0] : 14'b0;
//  assign        qb[13:0]= hact_b? do[13:0] : 14'b0;

  always @ (posedge wclk) if (we_ll) ll[10:0] <= d[10:0];

  always @ (posedge aclk) begin
    line_d  <= line;
    if      (!frame)          oddline <= 1'b1;
    else if (line && !line_d) oddline <= !oddline;
    if      (line && !line_d) oddpix  <= 1'b1;
    else                      oddpix  <= !oddpix;
  end

  merge_streams i_merge_streams(.clk(aclk),   // ADC clock, 25MHz
                        .clk2x(clk2x), // output clock, twice clk. 0->1 earlier than that of clk
                        .en(frame),    // enables operation (==frame)
                        .line(line),  // only leading edge is used
                        .ll(ll[10:0]),    // [10:0] - half line length.10 MSBs define (always even) line length to process, LSB - delay start by 1 pixel (to match 2 halves)
                        .da(dca[13:0]), // [13:0] channel A data in
                        .db(dcb[13:0]), // [13:0] channel B data in
                        .do(do[13:0]),  // [13:0] channel A data out (nothing to channel b)
                        .line_out(double_line)); // output line data valid

  window_proc i_wnd_a (.wclk(wclk), // interface clock
//                      .we(we && (a[3:2]==2'b10)),   // write enable (and cs)
                      .we(wew && (a[2]==1'b0)),   // write enable (and cs)
                      .a(a[1:0]),     // [1:0] register address 0..3 - left, width,top, height
                      .d(d[11:0]),    // [11:0]  - 12 bits
                      .aclk(clk2x),    // ADC/pixel clock
                      .frame(frame),  // high during frame 
                      .line(double_line),    // high during pixel shift 
                      .hact(hact_a),  // valid pixel output
                      .vact(vact_a)); // active frame output


/*
  window_proc i_wnd_b (.wclk(wclk), // interface clock
                      .we(wew && (a[2]==1'b1)),   // write enable (and cs)
                      .a(a[1:0]),     // [1:0] register address 0..3 - left, width,top, height
                      .d(d[11:0]),    // [11:0]  - 12 bits
                      .aclk(clk2x),    // ADC/pixel clock
                      .frame(frame),  // high during frame 
                      .line(double_line),    // high during pixel shift 
                      .hact(hact_b),  // valid pixel output
                      .vact(vact_b)); // active frame output
*/
  
  RAM16XnD # (16) i_gain_a(.D(d[15:0]),
                          .WE(weg && (a[3:2]==2'b10)),
                          .CLK(wclk),
                          .AW({2'b00,a[1:0]}),
                          .AR({2'b0,oddline,!oddpix}),
                          .QW(),
                          .QR(gain_a[15:0]));
  RAM16XnD # (16) i_gain_b(.D(d[15:0]),
                          .WE(weg && (a[3:2]==2'b11)),
                          .CLK(wclk),
                          .AW({2'b00,a[1:0]}),
                          .AR({2'b0,oddline,!oddpix}),
                          .QW(),
                          .QR(gain_b[15:0]));
  RAM16XnD # (14) i_sub_a(.D(d[13:0]),
                          .WE(weg && (a[3:2]==2'b00)),
                          .CLK(wclk),
                          .AW({2'b00,a[1:0]}),
                          .AR({2'b0,oddline,oddpix}),
                          .QW(),
                          .QR(sub_a[13:0]));
  RAM16XnD # (14) i_sub_b(.D(d[13:0]),
                          .WE(weg && (a[3:2]==2'b01)),
                          .CLK(wclk),
                          .AW({2'b00,a[1:0]}),
                          .AR({2'b0,oddline,oddpix}),
                          .QW(),
                          .QR(sub_b[13:0]));

endmodule

