/*
** -----------------------------------------------------------------------------**
** rtc333.v
**
** Real time clock, counts seconds (32 bit), microseconds (20 bits)
** allows fine adjustment (0.1 ppm to =/- 0.3%), used in timestamping images.
**
** Copyright (C) 2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: rtc324.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/

module rtc324    (clk125,   // 125MHz clock
                  sclk,    // system clock (pixel clock), posedge
                  exposure,// sensor exposure - start will be used to latch frame time stamp
                  we,      // sync to sclk                
                  wa,      // [2:0] write address
                           // 0 - write seconds high word - will write all other (actual write will happen after writing low word microseconds)
                           // 1 - write seconds low word (actual write will happen after writing low word microseconds)
                           // 2 - write microseconds high word (actual write will happen after writing low word microseconds)
                           // 3 - write microseconds low word and set the clock seconds, microseconds
                           // 4 - write correction (16 bit, signed
                           // 5 - nop, just latch the output 32+20 bits to read out).
                  wd,      // [15:0] data to write, valid 1 cycle after pre_we, wa
                  ssec,    // [31:0] latched system time, seconds
                  susec,   // [19:0] latched system time, usec -  0..999999 (0..0xf423f)
                  fsec,   // [19:0] latched exposure start, seconds
                  fusec);   // [31:0] latched exposure start, microseconds
// counter is updated each 0.5 (non-compensated) microsecond. Usually counters are incremented each other upd pulse, but an interval
// may be 1 or 3 upd periods (0.5/1.5 usec)
    input         clk125;
    input         sclk;
    input         exposure;
    input         we;
    input  [ 2:0] wa;
    input  [15:0] wd;
    output [31:0] ssec;
    output [19:0] susec;
    output [31:0] fsec;
    output [19:0] fusec;
    
// commands are sync to sclk (25MHz) posedge, but that can be changed with timing. halfusec is made sync to nearest sclk (provided sclk is fast enough)


    reg    [6:0]  usec_cntr=   0;
    reg           usec_cntr_rst=0;
    reg           halfusec125=0;
    reg           prehalfusec=0;
    reg           halfusec0=0;
    reg           halfusec=0;
    
    always @ (negedge clk125) begin // to match sclk (derived from posedge clk125)
      usec_cntr_rst  <= (usec_cntr[6:0]== 7'h7b); // period=125
      halfusec125 <= (usec_cntr[6:0]== 7'h3d) || usec_cntr_rst;
      usec_cntr[6:0] <= (usec_cntr_rst) ? 7'h0:(usec_cntr[6:0]+1);
    end
    // halfusec alternates between 0.496usec and 0.504usec
    // now sync to sclk
    always @ (negedge clk125 or posedge halfusec) if (halfusec) prehalfusec <= 1'b0; else if (halfusec125) prehalfusec <= 1'b1;
    always @ (posedge sclk) begin
      halfusec0 <= prehalfusec;
      halfusec  <= halfusec0 && !halfusec;
    end

    reg    [19:0] wusec;
    reg    [31:0] wsec;
    reg    [15:0] corr=0;
    
    
    always @ (posedge sclk) if (we) begin
      if (wa[2:0]==3'h0) wsec [31:16] <= wd[15:0];
      if (wa[2:0]==3'h1) wsec [15: 0] <= wd[15:0];
      if (wa[2:0]==3'h2) wusec[19:16] <= wd[ 3:0];
      if (wa[2:0]==3'h3) wusec[15: 0] <= wd[15:0];
      if (wa[2:0]==3'h4) corr [15: 0] <= wd[15:0];
    end    
    
    reg        set_clock;    // copy all parts of seconds and useconds registers written by CPU to the running clock
    reg        latch_clock;  // make a snapshot of the current clock to be read out
    reg        exp_start;    // copy currentr clock to "exposure start" registers
    reg  [1:0] expdly;       // used to generate exp_start
    reg [19:0] usec=0;    // running clock (microseconds)
    reg [31:0] sec=0;     // running clock (seconds)
    reg [31:0] ssec;
    reg [19:0] susec;
    reg [31:0] fsec;
    reg [19:0] fusec;
    reg        pre_inc_usec;
    reg        inc_usec;
    reg        inc_sec;
    wire       usec_last;
    assign     usec_last=(usec[19:0]==20'hf423f);
    
    reg   [23:0]  acc;
    wire  [24:0]  next_acc;
    assign next_acc[24:0]= {1'b0,acc[23:0]} + {1'b0,~corr [15], {7{corr [15]}}, corr[15:0]};
    
    always @ (posedge sclk) begin
      expdly[1:0] <= {expdly[0], exposure};
      exp_start   <= expdly[0] && !expdly[1];
      set_clock   <= we && (wa[2:0]==3'h3); 
      latch_clock <= we && (wa[2:0]==3'h5); 
      
      if (latch_clock) begin
        ssec [31:0] <= sec [31:0];
        susec[19:0] <= usec[19:0];
      end
      if (exp_start) begin
        fsec [31:0] <= sec [31:0];
        fusec[19:0] <= usec[19:0];
      end
      
      if      (set_clock) acc[23:0] <= 24'h0;
      else if (halfusec)  acc[23:0] <= next_acc[23:0];
      
      pre_inc_usec <= halfusec && next_acc[24];
    
      inc_usec    <= pre_inc_usec;
      inc_sec     <= pre_inc_usec && usec_last;
      
      if      (set_clock) usec[19:0] <= wusec[19:0];
      else if (inc_usec)  usec[19:0] <= (inc_sec)?20'h0:(usec[19:0] + 1);
      
      if      (set_clock) sec[31:0] <= wsec[31:0];
      else if (inc_sec)   sec[31:0] <= sec [31:0] + 1;
    
    end
    
    
endmodule


                             
