/*
** -----------------------------------------------------------------------------**
** timing.v
**
** Provides pixel-level timing, up to 16 c125 cycles (128ns), nominal  40ns
** Each transition is adjusted to 1/2 of the clock period (8/2 = 4 ns)
**
** Copyright (C) 2003-2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: timing.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/21 21:56:22  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.7  2007/08/21 21:56:22  elphel
**  Changed the duty cycle of the pixel clock from the 10347 board to the 10353. Added possibility to adjust the phase of that clock in 4ns steps (currently 40 degrees)
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.2  2006/09/24 16:45:51  elphel
**  added real time clock
**
*/

module timing (clk,	// global 125MHz input clock,
               we,   // enabable write timing
					a,    // address [4:0]
					      // 0 -  write period (default = 4 =>5 cycles, maximum 15->16 cycles). For foolproof everything less than 4 will add 4
							// 1 -  write PH inactive (for period ==5 should be 3)
							// 2 - ADCCLK active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 3 - ADCCLK active at (64-71)(68-75)(72-79) ... (120..127) ns
							// 4 - SHPA   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 5 - SHPA   active at (64-71)(68-75)(72-79) ... (120..127) ns
							// 6 - SHDA   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 7 - SHDA   active at (64-71)(68-75)(72-79) ... (120..127) ns
							// 8 - SHPB   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// 9 - SHPB   active at (64-71)(68-75)(72-79) ... (120..127) ns
							// a - SHDB   active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
							// b - SHDB   active at (64-71)(68-75)(72-79) ... (120..127) ns
                     // c - clk_adc active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) 
                     // d - clk_adc active at (64-71)(68-75)(72-79) ... (120..127) ns
                     // e - clk_2x  active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
                     // f - clk_2x  active at (64-71)(68-75)(72-79) ... (120..127) ns
                     //10 - clk_itf active at ( 0- 7)( 4-11)( 8-15) ... ( 56.. 63) ns
                     //11 - clk_itf  active at (64-71)(68-75)(72-79) ... (120..127) ns

					d,    // input data, [15:0]
//					sync, // 1 cycle long, just before cpix
					pre_sync, // (added one more ahead) 1 cycle long, just before cpix
					cpix, // global pixel clok input (was clk25), high  16ns
					pre_cpix, // 1 cycle before cpix (output)
					pre_rg, //1 cycle ahead of RG output, active high
					pre_ph, // 1 cycle ahead of PH1 going high, PH2 going low
					adcclk,
				   shpa,
				   shda,
				   shpb,
				   shdb,
					iclk_adc, // (before global) clock, normally sync to adcclk
					iclk2x,   // global clock, twice the clk_adc frequency
               iclkitf); // phase-shifted clk2x - to go to the main board

   input        clk,we;
	input  [4:0] a;
	input [15:0] d;
//	output       sync;
	output       pre_sync;
	input        cpix;
	output       pre_cpix;
	output       pre_rg;
	output       pre_ph;
	output		adcclk;
	output		shpa;
	output		shda;
	output		shpb;
	output		shdb;
	output		iclk_adc;
	output		iclk2x;
	output		iclkitf;

   wire    [3:0] pix_per;
	reg     [3:0] ph_inact_dur;  
	reg    [11:0] per_to_load;
//	reg           was_0;
	wire          was_0;
	wire   [13:0] clk_ph;

	wire          iclk_adc, iclk2x;
	reg           sync, sync_d;
	wire          pre_cpix= sync || sync_d;
	reg           pre_ph;
	wire          ph_on;
   wire          pre_rg=sync_d;
   wire        pre_sync=was_0;

	reg   [31:0] adcclk_dur;
	reg   [31:0] shpa_dur;
	reg   [31:0] shda_dur;
	reg   [31:0] shpb_dur;
	reg   [31:0] shdb_dur;
	reg   [31:0] clk_adc_dur;
	reg   [31:0] clk2x_dur;
	reg   [31:0] iclkitf_dur;
   always @ (posedge cpix) if (we) begin
	  if (a[4:0]== 5'h01) ph_inact_dur <=d[3:0];
     if (a[4:0]== 5'h02) adcclk_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h03) adcclk_dur[31:16] <= d[15:0]; 
     if (a[4:0]== 5'h04)   shpa_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h05)   shpa_dur[31:16] <= d[15:0]; 
     if (a[4:0]== 5'h06)   shda_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h07)   shda_dur[31:16] <= d[15:0]; 
     if (a[4:0]== 5'h08)   shpb_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h09)   shpb_dur[31:16] <= d[15:0]; 
     if (a[4:0]== 5'h0a)   shdb_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h0b)   shdb_dur[31:16] <= d[15:0]; 
     if (a[4:0]== 5'h0c)   clk_adc_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h0d)   clk_adc_dur[31:16] <= d[15:0]; 
     if (a[4:0]== 5'h0e)   clk2x_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h0f)   clk2x_dur[31:16] <= d[15:0]; 
     if (a[4:0]== 5'h10)   iclkitf_dur[15: 0] <= d[15:0]; 
     if (a[4:0]== 5'h11)   iclkitf_dur[31:16] <= d[15:0]; 
	end


	FDE i_pix_per0 (.C(cpix), .CE(we && (a[3:0]==4'h0)), .D(d[0]), .Q(pix_per[0]));
	FDE i_pix_per1 (.C(cpix), .CE(we && (a[3:0]==4'h0)), .D(d[1]), .Q(pix_per[1]));
	FDE i_pix_per2 (.C(cpix), .CE(we && (a[3:0]==4'h0)), .D(d[2] || !d[3]), .Q(pix_per[2]));
//synthesis translate_off
 defparam i_pix_per2.INIT = 1'b1;
//synthesis translate_on
//synthesis attribute INIT of i_pix_per2 is "1" 
	FDE i_pix_per3 (.C(cpix), .CE(we && (a[3:0]==4'h0)), .D(d[3]), .Q(pix_per[3]));
	always @ (posedge clk) begin
     case (pix_per)
	    4'hf:    per_to_load[11:0] <= 12'b100000000000; 
	    4'he:    per_to_load[11:0] <= 12'b010000000000; 
	    4'hd:    per_to_load[11:0] <= 12'b001000000000; 
	    4'hc:    per_to_load[11:0] <= 12'b000100000000; 
	    4'hb:    per_to_load[11:0] <= 12'b000010000000; 
	    4'ha:    per_to_load[11:0] <= 12'b000001000000; 
	    4'h9:    per_to_load[11:0] <= 12'b000000100000; 
	    4'h8:    per_to_load[11:0] <= 12'b000000010000; 
	    4'h7:    per_to_load[11:0] <= 12'b000000001000; 
	    4'h6:    per_to_load[11:0] <= 12'b000000000100; 
	    4'h5:    per_to_load[11:0] <= 12'b000000000010; 
	    default: per_to_load[11:0] <= 12'b000000000001;
 	  endcase
//	  was_0 <= !was_0 && (clk_ph[13:0]==14'b0);
	  sync     <= was_0;
	  sync_d <= sync;
//	  icpix    <= pre_cpix;
	end
  FD i_was_0  (.C(clk), .D(!was_0 && (clk_ph[13:0]==14'b0)), .Q(was_0));

  FD i_clk_ph0  (.C(clk), .D(was_0?1'b0:           clk_ph[ 1]), .Q(clk_ph[ 0]));
  FD i_clk_ph1  (.C(clk), .D(was_0?1'b0:           clk_ph[ 2]), .Q(clk_ph[ 1]));
  FD i_clk_ph2  (.C(clk), .D(was_0?per_to_load[ 0]:clk_ph[ 3]), .Q(clk_ph[ 2]));
  FD i_clk_ph3  (.C(clk), .D(was_0?per_to_load[ 1]:clk_ph[ 4]), .Q(clk_ph[ 3]));
  FD i_clk_ph4  (.C(clk), .D(was_0?per_to_load[ 2]:clk_ph[ 5]), .Q(clk_ph[ 4]));
  FD i_clk_ph5  (.C(clk), .D(was_0?per_to_load[ 3]:clk_ph[ 6]), .Q(clk_ph[ 5]));
  FD i_clk_ph6  (.C(clk), .D(was_0?per_to_load[ 4]:clk_ph[ 7]), .Q(clk_ph[ 6]));
  FD i_clk_ph7  (.C(clk), .D(was_0?per_to_load[ 5]:clk_ph[ 8]), .Q(clk_ph[ 7]));
  FD i_clk_ph8  (.C(clk), .D(was_0?per_to_load[ 6]:clk_ph[ 9]), .Q(clk_ph[ 8]));
  FD i_clk_ph9  (.C(clk), .D(was_0?per_to_load[ 7]:clk_ph[10]), .Q(clk_ph[ 9]));
  FD i_clk_ph10 (.C(clk), .D(was_0?per_to_load[ 8]:clk_ph[11]), .Q(clk_ph[10]));
  FD i_clk_ph11 (.C(clk), .D(was_0?per_to_load[ 9]:clk_ph[12]), .Q(clk_ph[11]));
  FD i_clk_ph12 (.C(clk), .D(was_0?per_to_load[10]:clk_ph[13]), .Q(clk_ph[12]));
  FD i_clk_ph13 (.C(clk), .D(was_0?per_to_load[11]:1'b0      ), .Q(clk_ph[13]));

//  BUFG   i_cpix (.I(icpix), .O(cpix));
  SRL16 i_ph_on (.CLK(clk), .Q(ph_on), .D(was_0),
                 .A0(ph_inact_dur[0]), .A1(ph_inact_dur[1]), .A2(ph_inact_dur[2]), .A3(ph_inact_dur[3]));
  always @ (posedge clk) begin
    if (sync)        pre_ph <= 1'b0;
	 else if (ph_on)  pre_ph <= 1'b1;
  end
 fine_time i_adcclk(.clk(clk),            // 125 MHz
                    .d(adcclk_dur[31:0]), //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),          // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),   // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(adcclk));          // output (inverted)
 fine_time i_shpa  (.clk(clk),            // 125 MHz
                    .d(shpa_dur[31:0]),   //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),          // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),   // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(shpa));            // output (inverted)
 fine_time i_shda  (.clk(clk),            // 125 MHz
                    .d(shda_dur[31:0]),   //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),          // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),   // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(shda));            // output (inverted)
 fine_time i_shpb  (.clk(clk),            // 125 MHz
                    .d(shpb_dur[31:0]),   //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),          // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),   // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(shpb));            // output (inverted)
 fine_time i_shdb  (.clk(clk),            // 125 MHz
                    .d(shdb_dur[31:0]),   //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),          // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),   // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(shdb));            // output (inverted)

 fine_time i_iclk_adc(.clk(clk),          // 125 MHz
                    .d(clk_adc_dur[31:0]),   //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),          // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),   // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(iclk_adc));        // output (inverted)

 fine_time i_iclk2x (.clk(clk),            // 125 MHz
                    .d(clk2x_dur[31:0]),   //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),           // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),    // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(iclk2x));           // output (inverted)
 fine_time i_iclkitf(.clk(clk),            // 125 MHz
                    .d(iclkitf_dur[31:0]), //[0] - active 8 ns just after sync (0-7ns), [1] - 4-11ns, [2] - 8-15ns,  d[31] - 120-127 ns
						  .sync(sync),           // 1-cycle long sequence sync (each 5-th cycle, sync to posedge)
						  .per(pix_per[3:0]),    // pixel period minus 1: 4 (period=5clk=40ns)...15 (period=16clk=128ns)
						  .q(iclkitf));          // output (inverted)
//  BUFG   i_clk_adc (.I(iclk_adc), .O(clk_adc));
//  BUFG   i_clk2x   (.I(iclk2x),   .O(clk2x)  );

endmodule
// reversed MSB/LSB
module fine_time(clk, // 125 MHz
                  d,   // [31:0]
							  // d[ 0] - active 8 ns just after sync (0-7ns),
							  // d[ 1] - 4-11ns,
							  // d[ 2] - 8-15ns,
							  // d[31] - 120-127 ns

                  per,  // pixel period minus one (4(default)->5,..15->16)

						sync,	// 1-cycle long sequence sync (one per pixel, sync to posedge)
						q);   // output (inverted)
   input       clk, sync;
	input [31:0] d;
   input [ 3:0] per;
	output q;
	wire   q;
   wire  sync_p; // "zero delay" copy of sync
	wire  sync_n; // sync to negedge;
   wire  syncp_p_pre, sync_p1; // separate siganl to fit into 4ns
   reg   [3:0] per_reg;
   reg   [3:0] per_m_1;
	reg [15:0] srp;
	reg [15:0] srn;
	always @ (posedge clk) begin
	  per_reg <= per;
     per_m_1 <= per-1;
	end
//  SRL16 i_sync_p     (.CLK(clk), .Q(sync_p),     .D(sync), .A0(1'b0), .A1(1'b0), .A2(1'b1), .A3(1'b0));
//  SRL16 i_sync_p_pre (.CLK(clk), .Q(sync_p_pre), .D(sync), .A0(1'b1), .A1(1'b1), .A2(1'b0), .A3(1'b0));
  SRL16 i_sync_p     (.CLK(clk), .Q(sync_p),     .D(sync), .A0(per_reg[0]), .A1(per_reg[1]), .A2(per_reg[2]), .A3(per_reg[3]));
  SRL16 i_sync_p_pre (.CLK(clk), .Q(sync_p_pre), .D(sync), .A0(per_m_1[0]), .A1(per_m_1[1]), .A2(per_m_1[2]), .A3(per_m_1[3]));
  FD    i_sync_p1    (.C(clk),   .Q(sync_p1), .D(sync_p_pre));
  FD_1  i_sync_n     (.C(clk),   .Q(sync_n),  .D(sync_p1));
   always @ (posedge clk) if (sync_p) srp[15:0] <= {d[30],d[28],d[26],d[24],d[22],d[20],d[18],d[16],
	                                                 d[14],d[12],d[10],d[ 8],d[ 6],d[ 4],d[ 2],d[ 0]};
	                       else        srp[15:0] <= {1'b0,srp[15:1]};
   always @ (negedge clk) if (sync_n) srn[15:0] <= {d[31],d[29],d[27],d[25],d[23],d[21],d[19],d[17],
	                                                 d[15],d[13],d[11],d[ 9],d[ 7],d[ 5],d[ 3],d[ 1]};
	                       else        srn[15:0] <= {1'b0,srn[15:1]};
   assign q=!srp[0] && !srn[0];
endmodule

